(* Compiler for Janus *)
(* Compile by mosmlc -c Compiler.sml *)
(* Then recompile JC by mosmlc JC.sml -o JC *)

structure Compiler :> Compiler =
struct

  (* Use "raise Error (message,position)" for error messages *)
  exception Error of string*(int*int)

  (* Name generator.  Call with, e.g., t1 = "tmp"^newName () *)
  val counter = ref 0

  fun newName () = (counter := !counter + 1;
                  "_" ^ Int.toString (!counter)^ "_")

  (* Number to text with spim-compatible sign symbol *)
  fun makeConst n = if n>=0 then Int.toString n
                    else "-" ^ Int.toString (~n)

  (* link register *)
  val RA = "31"
  (* Register for use in calls *)
  val FP = "30"
  (* Register for stack pointer *)
  val SP = "29"
  (* ADDED dedicated register for User Stack "Pointer" *)
  val USP = "28"

  (* Suggested register division *)
  val maxCaller = 9   (* highest caller-saves register *)
  val maxReg = 25     (* highest allocatable register *)
  (* this means Janus variables are primarily allocated in $10 - $25 *)
  (* and temporaries primarily in $2 ... $9 *)


  (* ADDED for IO: specific function for looking up variables in Def lists *)	
  fun lookupDef x []
        = NONE
    | lookupDef x (Janus.IntVarDef (y,pos)::table)
        = if x=y then SOME y else lookupDef x table
    | lookupDef x (Janus.ArrayVarDef (y,size,pos)::table)
        = if x=y then SOME y else lookupDef x table

  (* compile program *)
  fun compile (ins,outs,locals,stat,procs) =
    let
      val vars = findVars (ins@outs@locals)
      val readCode = makeReads ins
      val zeroCode = makeZero (outs@locals) ins  (* MODIFIED for IO: added ins as extra argument to makeZero *)
      val checkZeroCode = checkZero (ins@locals) outs  (* MODIFIED for IO: added outs as extra argument to checkZero *)
	  val checkEmptyStack =  (* ADDED for checking whether the stack is empty *)
		[Mips.BNE (USP,"0","_NotEmptyStack_")] (* error if stack is not empty *)
      val writeCode = makeWrites outs
      val procsCode = compileProcs procs vars
      val mainCode = compileStat stat vars
      val code = readCode @ zeroCode @ mainCode @ checkZeroCode @ checkEmptyStack @ writeCode  (* MODIFIED for stack *)
                 @ [Mips.J "_stop_"] @ procsCode
      val (code1, _, _)
             = RegAlloc.registerAlloc code [] 2 maxCaller maxReg
    in
      [Mips.TEXT "0x00400000",
       Mips.GLOBL "main",
       Mips.LABEL "main"]
		(* ADDED: initialise stack *)
	  @ [Mips.DATA "",     (* move to DATA area *)
	 	Mips.LABEL ("_stack_"), (* label of stack *)
	 	Mips.SPACE (makeConst (4*1000)), (* allocate space for 1000 elements *)
	 	Mips.TEXT "",  (* back to code area *)
		Mips.ADD (USP,"0","0")]     (* set USP to zero *)
      @ code1                     (* run program *)
      @ [Mips.LABEL "_stop_",
        Mips.LI ("2","10"),      (* syscall control = 10 *)
        Mips.SYSCALL,            (* exit *)
		(* BEGIN ADDED: code for reporting stack failures *)
		Mips.LABEL "_NotEmptyStack_",
		Mips.LA ("4","_NotEmptyStackString_"),
	 	Mips.LI ("2","4"), Mips.SYSCALL, (* print string *)
		Mips.LA ("4","_cr_"),
	 	Mips.LI ("2","4"), Mips.SYSCALL, (* print CR *)
	 	Mips.J "_stop_",
		Mips.LABEL "_PopNonZero_",
		Mips.LA ("4","_PopNonZeroString_"),
	 	Mips.LI ("2","4"), Mips.SYSCALL, (* print string *)
	 	Mips.MOVE ("4","5"),
	 	Mips.LI ("2","1"), Mips.SYSCALL, (* print line number *)
	 	Mips.LA ("4","_cr_"),
	 	Mips.LI ("2","4"), Mips.SYSCALL, (* print CR *)
	 	Mips.J "_stop_",
		Mips.LABEL "_PopEmpty_",
		Mips.LA ("4","_PopEmptyString_"),
	 	Mips.LI ("2","4"), Mips.SYSCALL, (* print string *)
	 	Mips.MOVE ("4","5"),
	 	Mips.LI ("2","1"), Mips.SYSCALL, (* print line number *)
	 	Mips.LA ("4","_cr_"),
	 	Mips.LI ("2","4"), Mips.SYSCALL, (* print CR *)
	 	Mips.J "_stop_",
		Mips.LABEL "_SwapError_",
		Mips.LA ("4","_SwapErrorString_"),
	 	Mips.LI ("2","4"), Mips.SYSCALL, (* print string *)
	 	Mips.MOVE ("4","5"),
	 	Mips.LI ("2","1"), Mips.SYSCALL, (* print line number *)
	 	Mips.LA ("4","_cr_"),
	 	Mips.LI ("2","4"), Mips.SYSCALL, (* print CR *)
	 	Mips.J "_stop_",
		(* END code for reporting stack failures *)
        Mips.LABEL "_Error_",    (* code for reporting assertion failures *)
	 	Mips.LA ("4","_ErrorString_"),
	 	Mips.LI ("2","4"), Mips.SYSCALL, (* print string *)
	 	Mips.MOVE ("4","5"),
	 	Mips.LI ("2","1"), Mips.SYSCALL, (* print line number *)
	 	Mips.LA ("4","_cr_"),
	 	Mips.LI ("2","4"), Mips.SYSCALL, (* print CR *)
	 	Mips.J "_stop_",
        Mips.LABEL "_NonZero_",  (* code for reporting nonzero variables *)
	 	Mips.LA ("4","_NonZeroString_"),
	 	Mips.LI ("2","4"), Mips.SYSCALL, (* print string *)
	 	Mips.J "_stop_",
        Mips.DATA "",
	 	Mips.LABEL "_cr_",       (* carriage return string *)
	 	Mips.ASCIIZ "\n",
	    (* ADDED stack error reporting strings *)
	    Mips.LABEL "_PopNonZeroString_",
	    Mips.ASCIIZ "ERROR: Cannot pop to non-zero variable, line ",
	    Mips.LABEL "_PopEmptyString_",
	    Mips.ASCIIZ "ERROR: Cannot pop from empty stack, line ",
		Mips.LABEL "_SwapErrorString_",
	    Mips.ASCIIZ "ERROR: Cannot swap with less than two elements on stack, line ",
		Mips.LABEL "_NotEmptyStackString_",
	    Mips.ASCIIZ "ERROR: Stack is not empty when ending program",
		(* END stack error reporting strings *)
	 	Mips.LABEL "_ErrorString_",
	 	Mips.ASCIIZ "ERROR: Assertion failed near line ",  (* MODIFIED slightly... *)
	 	Mips.LABEL "_NonZeroString_",
	 	Mips.ASCIIZ "ERROR: Nonzero non-output variable at end\n"]  (* MODIFIED slightly... *)
    end

  (* find all Janus variable names *)
  and findVars [] = []
    | findVars (Janus.IntVarDef (x,_)::defs) = x::(findVars defs)
    | findVars (Janus.ArrayVarDef (x,_,_)::defs) = x::(findVars defs)

  (* read input variables *)
  and makeReads [] = []
    | makeReads (Janus.IntVarDef (x,_)::defs) =
        [Mips.LI ("2","5"), (* read_int syscall *)
	 Mips.SYSCALL,
	 Mips.MOVE (x,"2")]
	@ makeReads defs
    | makeReads (Janus.ArrayVarDef (x,size,_)::defs) =
        [Mips.DATA "",     (* move to DATA area *)
	 Mips.LABEL ("_room_for_array_"^x), (* label of array *)
	 Mips.SPACE (makeConst (4*size)),
	 Mips.TEXT "",     (* back to code area *)
	 Mips.LA (x,"_room_for_array_"^x), (* load address of array *)
	 Mips.MOVE ("3","0"), (* offset := 0 *)
	 Mips.LI ("4",(makeConst (4*size))), (* end offset *)
	 Mips.LABEL ("_read_array_"^x),
	 Mips.LI ("2","5"), (* read_int syscall *)
	 Mips.SYSCALL,
	 Mips.ADD ("5",x,"3"), (* address of next element *)
	 Mips.SW ("2","5","0"), (* store element *)
	 Mips.ADDI ("3","3","4"), (* add 4 to offset *)
	 Mips.BNE ("3","4","_read_array_"^x)] (* until offset = end *)
	@ makeReads defs
	 
  (* initialize output variables to 0 *)
  and makeZero [] ins = []
    | makeZero (Janus.IntVarDef (x,_)::defs) ins =  (* MODIFIED for IO *)
		(case lookupDef x ins of
			NONE   => [Mips.MOVE (x,"0")] @ makeZero defs ins  (* zero normal out or local variable *)
		  | SOME _ => makeZero defs ins  (* skip if also an in variable *)
		)
    | makeZero (Janus.ArrayVarDef (x,size,_)::defs) ins =  (* MODIFIED for IO *)
		(case lookupDef x ins of 
			NONE   => (  (* zero normal out or local variable *)
        		[Mips.DATA "",     (* move to DATA area *)
	 			 Mips.LABEL ("_room_for_array_"^x), (* label of array *)
	 			 Mips.SPACE (makeConst (4*size)),
	 			 Mips.TEXT "",     (* back to code area *)
	 			 Mips.LA (x,"_room_for_array_"^x), (* load address of array *)
	 			 Mips.MOVE ("3","0"), (* offset := 0 *)
	 			 Mips.LI ("4",(makeConst (4*size))), (* end offset *)
	 			 Mips.LABEL ("_zero_array_"^x),
	 			 Mips.ADD ("5",x,"3"), (* address of next element *)
	 			 Mips.SW ("0","5","0"), (* store 0 to element *)
	 			 Mips.ADDI ("3","3","4"), (* add 4 to offset *)
	 			 Mips.BNE ("3","4","_zero_array_"^x)] (* until offset = end *)
				 @ makeZero defs ins )
		  |	SOME _ => makeZero defs ins  (* skip if also an in variable *)
		)
	 
  (* check that input variables are 0 *)
  and checkZero [] outs = []
    | checkZero (Janus.IntVarDef (x,_)::defs) outs = (* MODIFIED for IO *)
		(case lookupDef x outs of
			NONE   => (* check normal in or local variable *)
				[Mips.BNE (x,"0","_NonZero_")] @ checkZero defs outs
          | SOME _ => checkZero defs outs (* skip if also an out variable *)
		)
    | checkZero (Janus.ArrayVarDef (x,size,_)::defs) outs =  (* MODIFIED for IO *)
		(case lookupDef x outs of
			NONE   =>
        		[Mips.MOVE ("3","0"), (* offset := 0 *)
	 			 Mips.LI ("5",(makeConst (4*size))), (* end offset *)
	 			 Mips.LABEL ("_check_zero_array_"^x),
	 			 Mips.ADD ("4",x,"3"), (* address of next element *)
	 			 Mips.LW ("6","4","0"), (* read element *)
	 			 Mips.BNE("6","0","_NonZero_"), (* check if 0 *)
	 			 Mips.ADDI ("3","3","4"), (* add 4 to offset *)
	 			 Mips.BNE ("3","5","_check_zero_array_"^x)] (* until offset = end *)
				 @ checkZero defs outs
		  | SOME _ => checkZero defs outs  (* skip if also an out variable *)
		)

  (* write output variables *)
  and makeWrites [] = []
    | makeWrites (Janus.IntVarDef (x,_)::defs) =
        [Mips.MOVE ("4",x),
	 Mips.LI ("2","1"), (* write_int syscall *)
	 Mips.SYSCALL,
	 Mips.LA ("4","_cr_"),
	 Mips.LI ("2","4"), Mips.SYSCALL (* print CR *)]
	@ makeWrites defs
    | makeWrites (Janus.ArrayVarDef (x,size,_)::defs) =
        [Mips.MOVE ("3","0"), (* offset := 0 *)
	 Mips.LI ("6",(makeConst (4*size))), (* end offset *)
	 Mips.LABEL ("_write_array_"^x),
	 Mips.ADD ("5",x,"3"), (* address of next element *)
	 Mips.LW ("4","5","0"), (* load element *)
	 Mips.LI ("2","1"), (* write_int syscall *)
	 Mips.SYSCALL,
	 Mips.LA ("4","_cr_"),
	 Mips.LI ("2","4"), Mips.SYSCALL, (* print CR *)
	 Mips.ADDI ("3","3","4"), (* add 4 to offset *)
	 Mips.BNE ("3","6","_write_array_"^x)] (* until offset = end *)
	@ makeWrites defs
	 
  (* compile expression *)
  and compileExp e place =
    case e of
      Janus.Num (n,pos) =>
        if n<32768 then
	  		[Mips.LI (place, makeConst n)]
		else
	  		[Mips.LUI (place, makeConst (n div 32768)),
	   		 Mips.ORI (place, place, makeConst (n mod 32768))]
    | Janus.LVal (Janus.IntVar (x,p)) => [Mips.MOVE (place,x)]
    | Janus.LVal (Janus.ArrayIndex (x,e1,p)) =>
        let
	  		val t1 = "_index_"^newName()
          	val code1 = compileExp e1 t1
		in
	  		code1 @
	  		[Mips.SLL (t1,t1,"2"), (* scale index *)
	   		Mips.ADD (t1,t1,x), (* add array address *)
	   		Mips.LW (place,t1,"0")] (* load element *)
		end
    | Janus.Plus (e1,e2,pos) =>
        let
	  		val t1 = "_plus1_"^newName()
	  		val t2 = "_plus2_"^newName()
          	val code1 = compileExp e1 t1
          	val code2 = compileExp e2 t2
		in
	  		code1 @ code2 @ [Mips.ADD (place,t1,t2)]
		end
    | Janus.Minus (e1,e2,pos) =>
        let
	  		val t1 = "_minus1_"^newName()
	  		val t2 = "_minus2_"^newName()
          	val code1 = compileExp e1 t1
          	val code2 = compileExp e2 t2
		in
	  		code1 @ code2 @ [Mips.SUB (place,t1,t2)]
		end
    | Janus.Half (e1,pos) =>
        let
	  		val t1 = "_half_"^newName()
          	val code1 = compileExp e1 t1
		in
	  		code1 @ [Mips.SRA (place,t1,"1")]
		end


  (* compile condition *)
  and compileCond c tlab flab =
    case c of
      Janus.Equal (e1,e2,pos) =>
        let
	  		val t1 = "_eq1_"^newName()
	  		val t2 = "_eq2_"^newName()
          	val code1 = compileExp e1 t1
          	val code2 = compileExp e2 t2
		in
	  		code1 @ code2 @
	  		[Mips.BEQ (t1,t2,tlab), Mips.J flab]
		end
    | Janus.Less (e1,e2,pos) =>
        let
	  		val t1 = "_less1_"^newName()
	  		val t2 = "_less2_"^newName()
          	val code1 = compileExp e1 t1
          	val code2 = compileExp e2 t2
		in
	  		code1 @ code2 @
	  		[Mips.SLT (t1,t1,t2), Mips.BNE (t1,"0",tlab),Mips.J flab]
		end
    | Janus.Not (c1,pos) => compileCond c1 flab tlab
    | Janus.And (c1,c2,pos) => 
        let
	  		val l1 = "_and_"^newName()
          	val code1 = compileCond c1 l1 flab
          	val code2 = compileCond c2 tlab flab
		in
	  		code1 @ [Mips.LABEL l1] @ code2
		end
    | Janus.Or (c1,c2,pos) => 
        let
	  	  val l1 = "_or_"^newName()
          val code1 = compileCond c1 tlab l1
          val code2 = compileCond c2 tlab flab
		in
	  	  code1 @ [Mips.LABEL l1] @ code2
		end
	| Janus.Empty pos =>  (* ADDED for stack *)
		[Mips.BEQ (USP,"0",tlab), (* branch to tlab if USP = 0 *)
		 Mips.J flab ] (* jump to flab if USP != 0 *)

  (* compile statement *)
  and compileStat s vars =
    case s of
      Janus.Sequence (s1,s2,pos) =>
        compileStat s1 vars @ compileStat s2 vars
    | Janus.AddUpdate (Janus.IntVar (x,p),e1,pos) =>
        let
	  		val t1 = "_addupdate_"^newName()
	  		val code1 = compileExp e1 t1
		in
	  		code1 @ [Mips.ADD (x,x,t1)]
		end
    | Janus.AddUpdate (Janus.ArrayIndex (x,e1,p),e2,pos) =>
        let
	  		val t1 = "_index_"^newName()
	  		val t2 = "_addupdateA1_"^newName()
	  		val t3 = "_addupdateA2_"^newName()
          	val code1 = compileExp e1 t1
	  		val code2 = compileExp e2 t2
		in
	  		code1 @ code2 @
	  		[Mips.SLL (t1,t1,"2"), (* scale index *)
	   		Mips.ADD (t1,t1,x), (* add array address *)
	   		Mips.LW (t3,t1,"0"), (* load element *)
	   		Mips.ADD (t3,t3,t2), (* add value *)
	   		Mips.SW (t3,t1,"0")] (* store element *)
		end
    | Janus.SubUpdate (Janus.IntVar (x,p),e1,pos) =>
        let
	  		val t1 = "_subupdate_"^newName()
	  		val code1 = compileExp e1 t1
		in
	  		code1 @ [Mips.SUB (x,x,t1)]
		end
    | Janus.SubUpdate (Janus.ArrayIndex (x,e1,p),e2,pos) =>
        let
	  		val t1 = "_index_"^newName()
	  		val t2 = "_subupdateA1_"^newName()
	  		val t3 = "_subupdateA2_"^newName()
          	val code1 = compileExp e1 t1
	  		val code2 = compileExp e2 t2
		in
	  		code1 @ code2 @
	  		[Mips.SLL (t1,t1,"2"), (* scale index *)
	   		Mips.ADD (t1,t1,x), (* add array address *)
	   		Mips.LW (t3,t1,"0"), (* load element *)
	   		Mips.SUB (t3,t3,t2), (* subtract value *)
	   		Mips.SW (t3,t1,"0")] (* store element *)
		end
    | Janus.Skip pos => []
    | Janus.If (c1,s1,s2,c2,(line,col)) =>
        let
	  		val ltrue = "_if_t_"^newName()
	  		val lfalse = "_if_f_"^newName()
	  		val lok = "_fi_ok_"^newName()
	  		val lfail = "_fi_fail_"^newName()
      		val code1 = compileCond c1 ltrue lfalse
	  		val code2 = compileStat s1 vars
	  		val code3 = compileStat s2 vars
	  		val code4 = compileCond c2 lok lfail (* ok if true *)
	  		val code5 = compileCond c2 lfail lok (* ok if false *)
		in
	  		code1 @
	  		[Mips.LABEL ltrue] @ code2 @ code4 @
	  		[Mips.LABEL lfalse] @ code3 @ code5 @
	  		(* assertion failed: Report error *)
	  		[Mips.LABEL lfail, Mips.LI ("5",makeConst line),Mips.J "_Error_"] @
	  		(* normal exit from conditional *)
	  		[Mips.LABEL lok]
		end
    | Janus.Loop (c1,s1,s2,c2,(line,col)) =>
        let
	  		val lok = "_loop_ok_"^newName()
	  		val lfail = "_loop_fail_"^newName()
	  		val ltrue = "_loop_t_"^newName()
	  		val lfalse = "_loop_f_"^newName()
	  		val code1 = compileCond c1 lok lfail (* ok if true *)
	  		val code2 = compileStat s1 vars
            val code3 = compileCond c2 ltrue lfalse
	  		val code4 = compileStat s2 vars
	  		val code5 = compileCond c1 lfail lok (* ok if false *)
		in
	  		code1 @
	  		(* assertion failed: Report error *)
	  		[Mips.LABEL lfail, Mips.LI ("5",makeConst line),Mips.J "_Error_"] @
	  		(* assertion O.K: enter loop *)
	  		[Mips.LABEL lok] @ code2 @ code3 @
	  		(* if exit condition false, do s2, test assertion = false and loop *)
	  		[Mips.LABEL lfalse] @ code4 @ code5 @
	  		(* normal exit from loop *)
	  		[Mips.LABEL ltrue]
		end
    | Janus.Call (proc,pos) =>
        [Mips.ADDI (SP,SP,"-4"), (* push old return address *)
	 	 Mips.SW (RA,SP,"0"),
	 	 Mips.JAL ("_forward_"^proc,vars)] (* and JAL to procedure *)
		(* all global variables are preserved *)
    | Janus.Uncall (proc,pos) =>
        [Mips.ADDI (SP,SP,"-4"), (* push old return address *)
	 	 Mips.SW (RA,SP,"0"),
	 	 Mips.JAL ("_backward_"^proc,vars)] (* and JAL to inverse procedure *)
		(* all global variables are preserved *)
	| Janus.Push (Janus.IntVar (x,p),pos) =>  (* ADDED for stack *)
		[Mips.LA ("2","_stack_"), (* load address of stack *)
		 Mips.ADDI (USP,USP,"1"), (* increment USP by one *)
		 Mips.SLL ("3",USP,"2"), (* shift USP left by 2 *)
		 Mips.ADD ("2","2","3"), (* add stack base address and offset *)
		 Mips.MOVE ("3",x), (* store value of x in $3 *)
		 Mips.SW ("3","2","0"), (* store value of x on stack *)
		 Mips.MOVE (x,"0") ] (* set x to zero *)
	| Janus.Push (Janus.ArrayIndex (x,e,p),pos) =>  (* ADDED for stack *)
		let
	  		val t1 = "_index_"^newName()
          	val code1 = compileExp e t1
		in
	  		code1 @
	  		[Mips.SLL (t1,t1,"2"), (* scale index *)
	   		 Mips.ADD (t1,t1,x), (* add array address *)
	   		 Mips.LW ("4",t1,"0"), (* load x[e] *)
			 Mips.LA ("2","_stack_"), (* load address of stack *)
			 Mips.ADDI (USP,USP,"1"), (* increment USP by one *)
			 Mips.SLL ("3",USP,"2"), (* shift USP left by 2 *)
			 Mips.ADD ("2","2","3"), (* add stack base address and offset *)
			 Mips.SW ("4","2","0"), (* store value of x on stack *)
			 Mips.SW ("0",t1,"0") ] (* set x[e] to zero *)
		end
	| Janus.Pop (Janus.IntVar (x,p),(line,col)) =>  (* ADDED for stack *)
		[Mips.LI ("5",makeConst line),
		 Mips.BNE (x,"0","_PopNonZero_"), (* error if non-zero x *)
		 Mips.BEQ (USP,"0","_PopEmpty_"), (* error if stack is empty *)
		 Mips.LA ("2","_stack_"), (* load address of stack *)
		 Mips.SLL ("5",USP,"2"), (* shift USP left by 2 *)
		 Mips.ADD ("5","2","5"), (* add stack base address and offset *)
		 Mips.LW ("4","5","0"), (* load value at top of stack *)
		 Mips.MOVE (x,"4"), (* set x to value of top stack element *)
		 Mips.SW ("0","5","0"), (* zero top stack element *)
		 Mips.ADDI (USP,USP,"-1") ] (* decrement USP *)
	| Janus.Pop (Janus.ArrayIndex (x,e,p),(line,col)) =>  (* ADDED for stack *)
		let
	  		val t1 = "_index_"^newName()
          	val code1 = compileExp e t1
		in
	  		code1 @
	  		[Mips.SLL (t1,t1,"2"), (* scale index *)
	   		 Mips.ADD (t1,t1,x), (* add array address *)
	   		 Mips.LW ("4",t1,"0"), (* load x[e] *)
			 Mips.LI ("5",makeConst line),
			 Mips.BNE ("4","0","_PopNonZero_"), (* error if non-zero x[e] *)
			 Mips.BEQ (USP,"0","_PopEmpty_"), (* error if stack is empty *)
			 Mips.LA ("2","_stack_"), (* load address of stack *)
			 Mips.SLL ("5",USP,"2"), (* shift USP left by 2 *)
			 Mips.ADD ("5","2","5"), (* add stack base address and offset *)
			 Mips.LW ("4","5","0"), (* load value at top of stack *)
			 Mips.SW ("4",t1,"0"), (* set x[e] to value of top stack element *)
			 Mips.SW ("0","5","0"), (* zero top stack element *)
			 Mips.ADDI (USP,USP,"-1") ] (* decrement stack pointer *)
		end
	| Janus.Swap (line,col) =>  (* ADDED for stack *)
		[Mips.LA ("2","_stack_"), (* load address of stack *)
		 Mips.SLTI ("4",USP,"2"), (* set $4 = 1 if USP < 2 *)
		 Mips.LI ("5",makeConst line),
		 Mips.BNE ("4","0","_SwapError_"), (* error if USP < 2 *)
		 Mips.SLL ("4",USP,"2"), (* shift USP left by 2 *)
		 Mips.ADD ("5","2","4"), (* add stack base address and offset *)
		 Mips.LW ("6","5","0"), (* load value at top of stack *)
		 Mips.ADDI ("5","5","-4"), (* move to next stack element *)
		 Mips.LW ("7","5","0"), (* load second value from top of stack *)
		 Mips.SW ("6","5","0"), (* store first element value on second place *)
		 Mips.ADDI ("5","5","4"), (* move to top stack element *)
		 Mips.SW ("7","5","0") ] (* store second element value on top of stack *)

  and compileProcs [] vars = []
    | compileProcs ((pname,body,pos)::procs) vars =
        [Mips.LABEL ("_forward_"^pname)] @
		compileStat body vars @
		[Mips.MOVE (FP,RA), (* move return address to FP *)
	 	Mips.LW (RA,SP,"0"), (* restore old return address *)
	 	Mips.ADDI (SP,SP,"4"),
	 	Mips.JR (FP,vars)] (* and jump back *)
		(* all global variables are preserved *)
		@
		[Mips.LABEL ("_backward_"^pname)] @
		compileStat (Reverse body) vars @
		[Mips.MOVE (FP,RA), (* move return address to FP *)
	 	Mips.LW (RA,SP,"0"), (* restore old return address *)
	 	Mips.ADDI (SP,SP,"4"),
	 	Mips.JR (FP,vars)] (* and jump back *)
		(* all global variables are preserved *)
		@ compileProcs procs vars

  and Reverse (Janus.Sequence (s1,s2,pos)) =
        Janus.Sequence (Reverse s2, Reverse s1,pos)
    | Reverse (Janus.AddUpdate (lv,e,pos)) =
        Janus.SubUpdate (lv,e,pos)
    | Reverse (Janus.SubUpdate (lv,e,pos)) =
        Janus.AddUpdate (lv,e,pos)
    | Reverse (Janus.If (c1,s1,s2,c2,pos)) =
        Janus.If (c2,Reverse s1,Reverse s2,c1,pos)
    | Reverse (Janus.Loop (c1,s1,s2,c2,pos)) =
        Janus.Loop (c2,Reverse s1,Reverse s2,c1,pos)
    | Reverse (Janus.Call (proc,pos)) =
        Janus.Uncall (proc,pos)
    | Reverse (Janus.Uncall (proc,pos)) =
        Janus.Call (proc,pos)
    | Reverse (Janus.Skip pos) =
        Janus.Skip pos
	| Reverse (Janus.Push (lv,pos)) =  (* ADDED reverse for push *)
		Janus.Pop (lv,pos)
	| Reverse (Janus.Pop (lv,pos)) =  (* ADDED reverse for pop *)
		Janus.Push (lv,pos)
	| Reverse (Janus.Swap pos) =  (* ADDED reverse for swap *)
		Janus.Swap pos

end
